/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "globus_gridftp_server_s3_base.h"

static 
void *
globus_l_s3_send_thread(
    void                               *user_arg);

static
S3Status
globus_l_s3_send_properties(
    const S3ResponseProperties         *properties,
    void                               *callbackData);

static
void
globus_l_s3_send_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data);

static
S3Status
globus_l_s3_send_data(
    int                                 buffer_size,
    const char                         *buffer,
    void                               *callback_data);

static
S3Status
globus_l_s3_send_pending_buffers(
    globus_i_s3_transfer_t             *transfer_state);

static
bool
globus_l_s3_send_is_ignorable_range_error(
    S3Status                            status,
    const S3ErrorDetails               *error_details);

/**
 * @brief send
 * @details This interface function is called when the client requests to
 * receive a file from the server. The path name is interpreted as
 * /bucket/key
 *
 * This function attempts to get the object from s3 without checking if
 * it exists libs3 will give us an error to the globus_l_s3_send_complete()
 * function if there's a problem accessing the resource.
 */
void
globus_i_s3_send(
    globus_gfs_operation_t              op,
    globus_gfs_transfer_info_t         *transfer_info,
    void                               *user_arg)
{
    globus_i_s3_handle_t               *s3_handle = user_arg;
    globus_i_s3_transfer_t             *transfer_state = NULL;
    globus_thread_t                     t;
    int                                 rc;
    globus_result_t                     result = GLOBUS_SUCCESS;

    GlobusS3Enter();
    GlobusS3Info("send: op=%p pathname=%s\n",
                (void *) op,
                transfer_info->pathname);

    assert(s3_handle != NULL);

    transfer_state = malloc(sizeof(globus_i_s3_transfer_t) + strlen(transfer_info->pathname) + 1);
    if (transfer_state == NULL)
    {
        result = GlobusGFSErrorMemory("transfer_state");
        goto fail;
    }

    *transfer_state = (globus_i_s3_transfer_t)
    {
        .op                           = op,
        .transfer_info                = transfer_info,
        .pending_buffers_last         = &transfer_state->pending_buffers
    };
    globus_mutex_init(&transfer_state->mutex, NULL);
    globus_cond_init(&transfer_state->cond, NULL);
    strcpy(transfer_state->pathname, transfer_info->pathname);
    globus_i_s3_path_parse(transfer_state->pathname, &transfer_state->bucket_name, &transfer_state->key);

    transfer_state->bucket_context = GLOBUS_S3_BUCKET_CONTEXT(s3_handle, transfer_state->bucket_name);

    rc = globus_thread_create(
            &t,
            NULL,
            globus_l_s3_send_thread,
            transfer_state);

    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorGeneric("globus_thread_create");

        globus_cond_destroy(&transfer_state->cond);
        globus_mutex_destroy(&transfer_state->mutex);

fail:
        GlobusS3Info("finished_command: op=%p result=%#x\n",
                 (void *) transfer_state->op,
                 (unsigned int) result);
        globus_gridftp_server_finished_command( 
                op,
                result,
                NULL);

        free(transfer_state);
    }
    GlobusS3Exit();

    return;
}
/* globus_i_s3_send() */
            
static 
void *
globus_l_s3_send_thread(
    void                               *user_arg)
{
    globus_i_s3_transfer_t             *transfer_state = user_arg;
    globus_off_t                        start_byte = 0;
    globus_off_t                        byte_count = 0;
    globus_result_t                     result = GLOBUS_SUCCESS;

    GlobusS3Enter();

    globus_mutex_lock(&transfer_state->mutex);

    globus_gridftp_server_begin_transfer(transfer_state->op, 0, NULL);

    globus_gridftp_server_get_read_range(
            transfer_state->op,
            &start_byte,
            &byte_count);

    while (byte_count != 0 && transfer_state->result == GLOBUS_SUCCESS)
    {
        transfer_state->offset = start_byte;

        transfer_state->retries = 0;
        do
        {
            transfer_state->status = S3StatusOK;
            transfer_state->result = GLOBUS_SUCCESS;
            globus_i_s3_retry_delay(transfer_state->retries);
            globus_mutex_unlock(&transfer_state->mutex);
            /* Unlock here so that abort and read callbacks can be processed */

            GlobusS3Info("REST: op=%p S3_get_object(bucket=%s, key=%s) try=%d\n",
                            (void *) transfer_state->op,
                            transfer_state->bucket_name,
                            transfer_state->key,
                            transfer_state->retries);
            S3_get_object(
                    &transfer_state->bucket_context,
                    transfer_state->key,
                    NULL,
                    start_byte,
                    byte_count == -1 ? 0 : byte_count,
                    NULL,
                    &(S3GetObjectHandler)
                    {
                        .responseHandler =
                        {
                            .propertiesCallback = globus_l_s3_send_properties,
                            .completeCallback   = globus_l_s3_send_complete,
                        },
                        .getObjectDataCallback  = globus_l_s3_send_data
                    },
                    transfer_state);

            GlobusS3DebugOpStatus(S3_get_object, transfer_state);
        }
        while (S3_status_is_retryable(transfer_state->status)
                && (transfer_state->retries++) < globus_i_s3_max_retries);
        globus_mutex_lock(&transfer_state->mutex);

        globus_gridftp_server_get_read_range(
                transfer_state->op,
                &start_byte,
                &byte_count);

    }

    while (transfer_state->pending_buffers != NULL || transfer_state->registered_buffers != NULL)
    {
        globus_cond_wait(&transfer_state->cond, &transfer_state->mutex);
    }
    result = transfer_state->result;

    GlobusS3Info("finished_transfer: op=%p result=%#x\n",

                 (void *) transfer_state->op, 
                 (unsigned int) result);

    globus_gridftp_server_finished_transfer(transfer_state->op, result);

    globus_mutex_unlock(&transfer_state->mutex);

    globus_cond_destroy(&transfer_state->cond);
    globus_mutex_destroy(&transfer_state->mutex);
    free(transfer_state);

    GlobusS3Exit();

    return NULL;
}
/* globus_l_s3_send_thread() */

static
S3Status
globus_l_s3_send_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_data)
{
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    GlobusS3ExitStatus(status);
    return status;
}
/* globus_l_s3_send_properties() */

/**
 * @brief Send complete callback
 * @details
 * This function is called by libs3 when I poll the request_context for
 * a send. It is the final callback for the s3 request. Once this function
 * is called, we will receive no more data from s3 for this transfer, so we
 * can send the operation's current buffer even if it's not completely full.
 *
 * @todo Check to see if status is a restartable error---if so, restart up
 * to X times
 */
static
void
globus_l_s3_send_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_i_s3_transfer_t             *transfer_state = callback_data;

    GlobusS3Enter();

    assert(transfer_state != NULL);

    globus_mutex_lock(&transfer_state->mutex);
    if (status != S3StatusOK)
    {
        if (!globus_l_s3_send_is_ignorable_range_error(status, error_details))
        {
            goto out;
        }
        else
        {
            status = S3StatusOK;
        }
    }

    if (transfer_state->current_buffer != NULL)
    {
        *transfer_state->pending_buffers_last = transfer_state->current_buffer;
        transfer_state->pending_buffers_last = &transfer_state->current_buffer->next;

        transfer_state->current_buffer = NULL;
    }

    status = globus_l_s3_send_pending_buffers(transfer_state);

out:
    if (status != S3StatusOK)
    {
        transfer_state->result = GlobusGFSErrorS3(
                    status,
                    transfer_state->pathname,
                    error_details);
    }
    globus_mutex_unlock(&transfer_state->mutex);

    GlobusS3Exit();

    return;
}
/* globus_l_s3_send_complete() */

/**
 * @brief Send data callback
 * @details
 * This is called by libs3 to provide data from an S3 object. It is called
 * once for each data block. Since the buffer is reused by libs3 once we
 * return, we make a copy of the data into globus_i_s3_buffer_t blocks.
 *
 * After we have copied all of the data into one or more buffers, we call
 * globus_l_s3_send_pending_buffers() to process any full buffers.
 */
static
S3Status
globus_l_s3_send_data(
    int                                 buffer_size,
    const char                         *buffer,
    void                               *callback_data)
{
    globus_i_s3_transfer_t             *transfer_state = callback_data;
    size_t                              size = buffer_size;
    globus_i_s3_buffer_t *              current_buffer = NULL;
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    globus_mutex_lock(&transfer_state->mutex);
    while (size > 0)
    {
        size_t                          copy_size;

        current_buffer = globus_i_s3_buffer_get(transfer_state, size);

        if (current_buffer == NULL)
        {
            transfer_state->result = GlobusGFSErrorMemory("buffer");

            status = S3StatusOutOfMemory;
            goto out;
        }

        if (size > (current_buffer->buffer_len - current_buffer->buffer_used))
        {
            copy_size = current_buffer->buffer_len
                      - current_buffer->buffer_used;
        }
        else
        {
            copy_size = size;
        }

        assert(copy_size > 0);

        memcpy(current_buffer->buffer + current_buffer->buffer_used,
               buffer,
               copy_size);

        current_buffer->buffer_used += copy_size;
        buffer += copy_size;
        size -= copy_size;

        if (current_buffer->buffer_used == current_buffer->buffer_len)
        {
            *transfer_state->pending_buffers_last = current_buffer;
            transfer_state->pending_buffers_last = &current_buffer->next;
            transfer_state->current_buffer = NULL;
        }
    }

    status = globus_l_s3_send_pending_buffers(transfer_state);


out:
    globus_mutex_unlock(&transfer_state->mutex);
    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_send_data() */

/**
 * @brief Send data write callback
 * @details
 * This is called by the gridftp server once it has sent the data block
 * passed to globus_gridftp_server_register_write(). Since this callback
 * isn't triggered by s3 polling, we must hold the lock before doing
 * anything here. We put the used buffer back into the free list, and then
 * check to see if the operation has completed and potentially finish
 * the operation.
 */
static
void
globus_l_s3_send_write_callback(
    globus_gfs_operation_t              op,
    globus_result_t                     result,
    globus_byte_t                      *buffer,
    globus_size_t                       nbytes,
    void                               *user_arg)
{
    globus_i_s3_transfer_t             *transfer_state = user_arg;
    globus_i_s3_buffer_t               *ptr, **last_next;

    GlobusS3Enter();

    globus_mutex_lock(&transfer_state->mutex);

    ptr = transfer_state->registered_buffers;
    last_next = &transfer_state->registered_buffers;

    while (ptr != NULL)
    {
        if (ptr->buffer == buffer)
        {
            /* Clear used offset and return to the free_buffers list */
            ptr->buffer_used = 0;
            *last_next = ptr->next;
            ptr->next = transfer_state->free_buffers;
            transfer_state->free_buffers = ptr;
            break;
        }
        last_next = &ptr->next;
        ptr = ptr->next;
    }

    if (transfer_state->pending_buffers == NULL
        && transfer_state->registered_buffers == NULL)
    {
        globus_cond_signal(&transfer_state->cond);
    }

    globus_mutex_unlock(&transfer_state->mutex);

    GlobusS3Exit();
}
/* globus_l_s3_send_write_callback() */

/**
 * @brief Send pending buffers
 * @details
 * This function call globus_gridftp_server_register_write() for each
 * complete buffer in this handle, storing pointers to the buffers in the 
 * registered_buffers list. If this returns an error status, the transfer
 * will be canceled.
 * @note call with transfer_info locked
 */
static
S3Status
globus_l_s3_send_pending_buffers(
    globus_i_s3_transfer_t             *transfer_state)
{
    S3Status                            status = S3StatusOK;
    globus_result_t                     result = GLOBUS_SUCCESS;

    GlobusS3Enter();

    while (transfer_state->pending_buffers)
    {
        globus_i_s3_buffer_t           *buffer;

        buffer = transfer_state->pending_buffers;

        GlobusS3Debug("registering write buffer=%p size=%zd offset=%"PRIuMAX"\n",
                buffer->buffer,
                buffer->buffer_used,
                (uintmax_t) transfer_state->offset);

        result = globus_gridftp_server_register_write(
                transfer_state->op,
                buffer->buffer,
                buffer->buffer_used,
                transfer_state->offset,
                -1,
                globus_l_s3_send_write_callback,
                transfer_state);

        if (result != GLOBUS_SUCCESS)
        {
            transfer_state->result = result;
            status = S3StatusInternalError;
            goto out;
        }
        transfer_state->offset += buffer->buffer_used;
        transfer_state->pending_buffers = buffer->next;

        if (transfer_state->pending_buffers == NULL)
        {
            transfer_state->pending_buffers_last = &transfer_state->pending_buffers;
        }
        buffer->next = transfer_state->registered_buffers;
        transfer_state->registered_buffers = buffer;
    }
out:
    GlobusS3ExitStatus(status);
    return status;
}
/* globus_l_s3_send_pending_buffers() */

static
bool
globus_l_s3_send_is_ignorable_range_error(
    S3Status                            status,
    const S3ErrorDetails               *error_details)
{
    intmax_t                            range_requested_start=0;
    intmax_t                            range_requested_end=0;
    intmax_t                            actual_object_size=-1;

    if (status != S3StatusErrorInvalidRange)
    {
        return false;
    }
    for (int i = 0; i < error_details->extraDetailsCount; i++)
    {
        const char                     *name = error_details->extraDetails[i].name;
        const char                     *value = error_details->extraDetails[i].value;

        if (strcmp(name, "RangeRequested") == 0)
        {
            if (sscanf(value, "bytes=%"SCNiMAX"-%"SCNiMAX"",
                        &range_requested_start,
                        &range_requested_end) <= 0)
            {
                return false;
            }
        }
        else if (strcmp(name, "ActualObjectSize") == 0)
        {
            if (sscanf(value, "%"SCNiMAX, &actual_object_size) <= 0)
            {
                return false;
            }
        }
    }
    /* Ignore this error if we asked for range bytes=${actual_object_size}- */
    if (range_requested_start != 0
        && range_requested_start == actual_object_size
        && range_requested_end == 0)
    {
        return true;

    }
    return false;
}
/* globus_l_s3_send_is_ignorable_range_error() */
    
// vim: ft=c:
